/* Prova d'esame 05/12/2006 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define DIM 100
#define N 10
#define K 5
#define M 5

struct sockaddr_in local;
struct sockaddr_in remote;


/********************************************************/


void gestore(int signo)
{
  int stato;
  printf("esecuzione gestore di SIGCHLD\n");
  wait(&stato);
}


/********************************************************/


main(int argc,char** argv)
{
  int letti, n, livello,nread,j, i = 0, len = sizeof(struct sockaddr_in);
  /* buffer deve contenere utenti e risorse di un livello piu' eventuali
   * separatori: due '\t' fra i nomi degli utenti e il terminatore finale
   *
   * livelliLiberati deve contenere i nomi dei livelli e N '\n', piu' il terminatore
   * finale
   */
  char scelta,buffer[((K+M)*DIM)+((K+M)*2)+1];

  // Matrice
  char s[N][K+M][DIM];

  int  listen_sd, conn_sd;
  int port;
  const int on = 1;
  struct sockaddr_in cliaddr, servaddr;

  /* CONTROLLO ARGOMENTI ---------------------------------- */
  if(argc!=2)
  {
    printf("Error: %s port\n", argv[0]);
    exit(1);
  }
  else port = atoi(argv[1]);

  /* INIZIALIZZAZIONE STRUTTURA DATI ----------------------------------------- */
  for(i=0; i<N; i++)
    for(j=0;j<K+M; j++)
      strcpy(s[i][j],"L");

  // Riempimento alcune posizioni per prova
  strcpy(s[0][0], "Aragorn");
  strcpy(s[0][K - 1], "Aron");
  strcpy(s[0][K], "Borraccia");
  strcpy(s[0][K + M - 1], "Arco");
  strcpy(s[1][0], "Frodo");
  strcpy(s[1][1], "Pipino");
  strcpy(s[1][K], "Anello");
  strcpy(s[1][K + 1], "Faretra");
  strcpy(s[N - 1][1], "Gandalf");
  strcpy(s[N - 1][K], "Bastone");


  /* INIZIALIZZAZIONE INDIRIZZO SERVER ----------------------------------------- */

  memset ((char *)&servaddr, 0, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = INADDR_ANY;
  servaddr.sin_port = htons(port);


  /* CREAZIONE E SETTAGGI SOCKET D'ASCOLTO --------------------------------------- */

  listen_sd=socket(AF_INET, SOCK_STREAM, 0);
  if(listen_sd <0)
  {perror("creazione socket "); exit(1);}
  printf("Server: creata la socket d'ascolto per le richieste di ordinamento, fd=%d\n", listen_sd);

  if(bind(listen_sd,(struct sockaddr *) &servaddr, sizeof(servaddr))<0)
  {perror("bind socket d'ascolto"); exit(1);}
  printf("Server: bind socket d'ascolto ok\n");

  if(setsockopt(listen_sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0)
  {perror("set opzioni socket d'ascolto"); exit(1);}
  printf("Server: set opzioni socket d'ascolto ok\n");

  if (listen(listen_sd, 5)<0)
  {perror("listen"); exit(1);}
  printf("Server: listen ok\n");


  /* AGGANCIO GESTORE PER EVITARE FIGLI ZOMBIE,
   * Quali altre primitive potrei usare? E' portabile su tutti i sistemi?
   * Pregi/Difetti?
   * Alcune risposte le potete trovare nel materiale aggiuntivo!
   */
  signal(SIGCHLD, gestore);


  /* CICLO DI RICEZIONE RICHIESTE --------------------------------------------- */
  for(;;)
  {
  	len = sizeof(cliaddr);
    if((conn_sd=accept(listen_sd,(struct sockaddr *)&cliaddr,&len))<0)
    {
      if (errno==EINTR)
      {
        perror("Forzo la continuazione della accept");
        continue;
      }
      else exit(1);
    }

    if(fork() == 0)
    {
	  /*
	   * NOTA: la struttura dati viene copiata nello
	   * spazio del memoria del processo figlio, e tutte
	   * le eventuali modifiche fatte dal figlio vengono perse
	   * dal padre.
	   *
	   * Nella soluzione qui proposta andiamo nella direzione della
	   * semplificazione e assumiamo che il figlio faccia solo
	   * operazioni di lettura per concentrarci sugli aspetti di
	   * comunicazione.
	   *
	   * Possibili soluzioni al problema sopra sono l'uso di un file di
	   * appoggio per la comunicazione inter-processo fra padre e figlio
	   * con letture e scritture della struttura dati da tale file, oppure
	   * l'utilizzo di uno spazio di memoria condiviso fra padre e figlio.
	   *
	   * Chiaramente, se il server fosse sequenziale e non paralello questi
	   * problemi non sussisterebbero perche' avremmo UN SOLO processo che
	   * accede in scrittura alla struttura dati.
	   */


      printf("\nFiglio: inizio!!");
      close(listen_sd);
      while( (nread = read(conn_sd, &scelta, 1)) > 0)
      {
        /*
         * VISUALIZZAZIONE:
         * Restituisco la lista degli utenti e delle risorse
         * eliminando le posizioni libere.
         *
         * Come utile esercizio si provi a spezza l'invio in due invii
         * consecutivi, prima si inviano gli utenti, poi le risorse.
         */

        if(scelta == 'v')
        {
          if(read(conn_sd,&livello,sizeof(int)) < 0)
          {
            perror("\nProblemi in lettura, chiudo il figlio");
            exit(1);  //sono nel figlio
          }
          livello=ntohl(livello);
          printf("\nRichiesto livello: %d\n",livello);

          strcpy(buffer, "");
          // Scrittura del livello richiesto
          for(i=0; i<K+M; i++){
          	if( strcmp(s[livello][i], "L")!=0 ){
          		strcat(buffer, s[livello][i]);
          		strcat(buffer, "\t\t");
          	}
          }
          printf("Livello da inviare:\n%s\n", buffer);


          i=strlen(buffer);
          i=htonl(i);
          /*
           * Caso piu' ampio e generale possibile.
           * Prima invio lunghezza dell'array di caratteri che devo
           * scrivere al pari, poi scrittura dell'array
           *
           * NOTA: spesso le cose sono semplificate, si evita
           * cioe' l'invio della lunghezza ipotizzando che per
           * invii brevi non ci siano problemi.
           */
          if( write(conn_sd,&i,sizeof(int)) < 0 )
          {
            perror("\nProblemi in scrittura, chiudo!");
            exit(1);
          }

          /*
           * Scrittura della riga su socket con singolo invio.
           * Alternativa possibile, invii multipli.
           *
           * Per l'invio di quantita' di dati maggiori, ad
           * esempio file, si usino sempre invii multipli!
           *
           */
          i=write(conn_sd, buffer, strlen(buffer));
          if ( (i<0) || (i!=strlen(buffer)) )
          {
            perror("\nProblemi in scrittura, chiudo il figlio");
            exit(1);  //sono nel figlio
          }

        }

        // ELIMINAZIONE
        if(scelta == 'e')
        {
          /*
           * Caso piu' ampio e generale possibile.
           * Prima lettura lunghezza dell'invio del pari,
           * poi lettura dello stream fino alla lunghezza.
           *
           * NOTA: spesso le cose sono semplificate, si evita
           * cioe' l'invio della lunghezza ipotizzando che per
           * invii brevi non ci siano problemi.
           */
          if( read(conn_sd,&n,sizeof(int)) < 0 )
          {
            perror("\nProblemi in lettura, chiudo!");
            exit(1);
          }

          // Converto la lunghezza della stringa letta
          n=ntohl(n);
          letti=0;
          while( letti< n ){
            /*
          	 * Aggiungo i caratteri letti alla fine dell'array
         	 * di caratteri e leggo i caratteri rimanenti.
      	     * Se tutto va bene effettuo una sola lettura di
         	 * n caratteri
         	 * NOTA: buffer e' un puntatore a carattere;
         	 * buffer+letti, e' il puntatore al prossimo
         	 * elemento dell'array di caratteri da riempire
         	 *
         	 */
      	    i=read(conn_sd, buffer+letti, n-letti);
      	    if(i<0)
            {
              perror("\nProblemi in lettura, chiudo!");
              exit(1);
            }
            letti=letti+i;
          }
          // Aggiungo il terminatore finale
          buffer[letti]='\0';


		  // Invio la lista dei livelli, -1 indica la terminazione degli invii
          for(i=0; i<N; i++)
            for(j=K-1; j<K+M; j++)
              if( strcmp(s[i][j], buffer)==0 ){
              	// Libero la risorsa inserendo una 'L'
              	strcpy(s[i][j], "L");
              	// Invio l'identificatore del livello liberato
              	i=htonl(i);
              	if(write(conn_sd, &i, sizeof(int))<0)
                {
                  perror("\nProblemi in scrittura, chiudo!");
                  exit(1);
                }
                i=ntohl(i);
              }
          i=-1;
          i=htonl(i);
          write(conn_sd, &i, sizeof(int));
        }

      } // while
      if(nread < 0)
      {
        perror("\nImpossibile leggere scelta: termino..");
        exit(1);
      }
      else close(conn_sd); //Se arrivo qui nread==0 -> raggiunta EOF, quindi chiudo
      printf("\nTermino il figlio!\n");
      exit(0);  // sono nel figlio
    }

    close(conn_sd); // sono nel padre
  }  // for(;;)
}

